#!/usr/bin/env python3
#
# Tests the methods and classes in myokit.tools
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import re
import sys
import threading
import time
import unittest

import myokit

from myokit.tests import TemporaryDirectory


class BenchmarkerTest(unittest.TestCase):
    """Tests the ``Benchmarker``."""

    def test_time(self):
        # Test benchmarker.time()

        b = myokit.tools.Benchmarker()
        x = [0] * 1000
        t0 = b.time()
        self.assertTrue(t0 >= 0)
        x = [0] * 1000
        t1 = b.time()
        self.assertTrue(t1 >= t0)
        x = [0] * 1000
        t2 = b.time()
        self.assertTrue(t2 >= t1)
        for i in range(1000):
            x = [0] * 1000
        t3 = b.time()
        self.assertTrue(t3 >= t2)
        b.reset()
        t4 = b.time()
        self.assertTrue(t4 < t3)

    def test_format(self):
        # Test benchmarker.format()

        b = myokit.tools.Benchmarker()
        self.assertEqual(b.format(1), '1 second')
        self.assertEqual(b.format(61), '1 minute, 1 second')
        self.assertEqual(b.format(60), '1 minute, 0 seconds')
        self.assertEqual(b.format(180), '3 minutes, 0 seconds')
        self.assertEqual(b.format(3600), '1 hour, 0 minutes, 0 seconds')
        self.assertEqual(b.format(3661), '1 hour, 1 minute, 1 second')
        self.assertEqual(
            b.format(3600 * 24), '1 day, 0 hours, 0 minutes, 0 seconds')
        self.assertEqual(
            b.format(3600 * 24 * 7),
            '1 week, 0 days, 0 hours, 0 minutes, 0 seconds')
        self.assertEqual(b.format()[-8:], ' seconds')

    def test_print(self):
        # Test benchmarker.print

        messages = [
            'Hello',
            'Yes yes',
            'Line 3',
        ]
        with myokit.tools.capture() as c:
            b = myokit.tools.Benchmarker()
            self.assertEqual(c.out(), '')
            self.assertEqual(c.err(), '')
            for m in messages:
                b.print(m)
        self.assertEqual(c.err(), '')
        lines = c.out().splitlines()
        self.assertEqual(len(lines), 3)

        r = re.compile(r'\[[ 0-9]+ us \([ 0-9]+ us\)\] ([ a-zA-Z0-9]+)')
        for line, msg in zip(lines, messages):
            m = r.match(line)
            self.assertIsNotNone(m)
            self.assertEqual(m.group(1), msg)


class CaptureTest(unittest.TestCase):
    """Test the ``capture`` context manager."""

    def test_capture_disabled(self):
        # Tests creating a capture manager that doesn't capture.

        with myokit.tools.capture(enabled=True) as p:
            with myokit.tools.capture(enabled=False) as q:
                print('2', end='')
                print('b', end='', file=sys.stderr)
                print('f', end='', file=sys.stderr)
            print('7', end='')
            print('g', end='', file=sys.stderr)

        self.assertEqual(p.out(), '27')
        self.assertEqual(p.err(), 'bfg')
        self.assertEqual(q.out(), '')
        self.assertEqual(q.err(), '')

    def test_capture_nested(self):
        # Tests capturing in a nested pattern.
        r = myokit.tools.capture(False)
        q = myokit.tools.capture(True)
        self.assertEqual(r.out(), '')
        self.assertEqual(r.err(), '')
        with myokit.tools.capture(False) as p:
            self.assertEqual(p.out(), '')
            self.assertEqual(p.err(), '')
            print('1', end='')
            print('a', end='', file=sys.stderr)
            with q:
                print('2', end='')
                print('b', end='', file=sys.stderr)
                with r:
                    print('3', end='')
                    print('c', end='', file=sys.stderr)
                    with myokit.tools.capture(True) as s:
                        print('4', end='')
                        print('d', end='', file=sys.stderr)
                    print('5', end='')
                    print('e', end='', file=sys.stderr)
                print('6', end='')
                print('f', end='', file=sys.stderr)
            print('7', end='')
            print('g', end='', file=sys.stderr)

        # Check captured text
        self.assertEqual(p.out(), '17')
        self.assertEqual(p.err(), 'ag')
        self.assertEqual(p.text(), '17ag')

        self.assertEqual(q.out(), '26')
        self.assertEqual(q.err(), 'bf')
        self.assertEqual(q.text(), '26bf')

        self.assertEqual(r.out(), '35')
        self.assertEqual(r.err(), 'ce')
        self.assertEqual(r.text(), '35ce')

        self.assertEqual(s.out(), '4')
        self.assertEqual(s.err(), 'd')
        self.assertEqual(s.text(), '4d')

    def test_capture_repeated_use(self):
        # Tests capturing in a nested pattern with repeated enters/exits.
        x = myokit.tools.capture()
        y = myokit.tools.capture(True)
        self.assertEqual(x.out(), '')
        with x:
            print('1', end='')
            print('a', end='', file=sys.stderr)
            self.assertEqual(x.out(), '')
            with y:
                print('2', end='')
                print('b', end='', file=sys.stderr)
                with x:
                    print('3', end='')
                    print('c', end='', file=sys.stderr)
                    with y:
                        print('4', end='')
                        print('d', end='', file=sys.stderr)
                    print('5', end='')
                    print('e', end='', file=sys.stderr)
                print('6', end='')
                print('f', end='', file=sys.stderr)
            print('7', end='')
            print('g', end='', file=sys.stderr)

        # Check captured text
        self.assertEqual(x.out(), '17')
        self.assertEqual(x.err(), 'ag')
        self.assertEqual(x.text(), '17ag')

        self.assertEqual(y.out(), '23456')
        self.assertEqual(y.err(), 'bcdef')
        self.assertEqual(y.text(), '23456bcdef')

        with x:
            print('hey')
            print('ya', file=sys.stderr)
        self.assertEqual(x.out(), 'hey\n')
        self.assertEqual(x.err(), 'ya\n')

        with y:
            print('foo')
            print('bar', file=sys.stderr)
        self.assertEqual(y.out(), 'foo\n')
        self.assertEqual(y.err(), 'bar\n')

    def test_capture_with_threads(self):
        # Tests capturing with threading.

        # Sleep times for each thread. Duration doesn't matter much, as will be
        # executed sequentially. But choose so that interlacing would occur if
        # not.
        times = [0.2, 0.01, 0.05]
        captured = [None, None, None]

        # Function to call inside threads
        def f(i):
            with myokit.tools.capture() as c:
                print(str(i) + 'a ', end='')
                time.sleep(times[i])
                print(str(i) + 'b ', end='')
                print(str(i) + 'e ', end='', file=sys.stderr)
            captured[i] = c.text().strip()

        # Call function from threads
        ps = []
        for i in range(len(times)):
            p = threading.Thread(target=f, args=(i,))
            ps.append(p)
        for p in ps:
            p.start()
            time.sleep(0.01)
        for p in ps:
            if p.is_alive():
                p.join()

        # Check captured output
        self.assertEqual(len(captured), 3)
        self.assertEqual(captured[0], '0a 0b 0e')
        self.assertEqual(captured[1], '1a 1b 1e')
        self.assertEqual(captured[2], '2a 2b 2e')


class ToolsTest(unittest.TestCase):
    """Tests various tools in myokit.tools"""

    def test_format_path(self):
        # Test format_path().
        fp = myokit.tools.format_path

        # Normal use
        self.assertEqual(
            fp(os.path.join('a', 'b', 'c')), os.path.join('a', 'b', 'c'))

        # No trailing slash
        self.assertEqual(fp('a'), 'a')
        self.assertEqual(fp('a/b/'), os.path.join('a', 'b'))

        # Use with custom root
        root = os.path.join(os.path.abspath('.'), 'a')
        self.assertEqual(
            fp(os.path.join(root, 'b', 'c'), root), os.path.join('b', 'c'))

        # Empty path
        self.assertEqual(fp(''), '.')
        self.assertEqual(fp('.'), '.')

        # Filesystem root
        self.assertEqual(fp('/'), os.path.abspath('/'))
        self.assertEqual(fp('/', root='/'), '.')

        # Path outside of root
        self.assertEqual(
            fp(os.path.abspath('test'), os.path.abspath('test/tost')),
            os.path.abspath('test'))

    def test_levenshtein_distance(self):
        # Test the levenshtein distance method.
        self.assertEqual(myokit.tools.lvsd('kitten', 'sitting'), 3)
        self.assertEqual(myokit.tools.lvsd('sitting', 'kitten'), 3)
        self.assertEqual(myokit.tools.lvsd('saturday', 'sunday'), 3)
        self.assertEqual(myokit.tools.lvsd('sunday', 'saturday'), 3)
        self.assertEqual(myokit.tools.lvsd('michael', 'jennifer'), 7)
        self.assertEqual(myokit.tools.lvsd('jennifer', 'michael'), 7)
        self.assertEqual(myokit.tools.lvsd('jennifer', ''), 8)
        self.assertEqual(myokit.tools.lvsd('', 'jennifer'), 8)
        self.assertEqual(myokit.tools.lvsd('', ''), 0)

    def test_natural_sort_key(self):
        # Test natural sort key method.

        a = ['a12', 'a3', 'a11', 'a2', 'a10', 'a1']
        b = ['a1', 'a2', 'a3', 'a10', 'a11', 'a12']
        self.assertNotEqual(a, b)
        a.sort()
        self.assertNotEqual(a, b)
        a.sort(key=myokit.tools.natural_sort_key)
        self.assertEqual(a, b)

    def test_rmtree(self):
        # Test rmtree

        with TemporaryDirectory() as d:

            # Create dir with subdir
            path = d.path('a')
            os.mkdir(path)
            os.mkdir(os.path.join(path, 'b'))

            # This can't be deleted with rmdir
            self.assertRaises(OSError, os.rmdir, path)
            self.assertTrue(os.path.exists(path))

            # But rmtree works
            myokit.tools.rmtree(path)
            self.assertFalse(os.path.exists(path))

            # It doesn't work twice in a row:
            self.assertRaises(OSError, myokit.tools.rmtree, path)

            # But we can ignore the exception
            myokit.tools.rmtree(path, silent=True)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Component and VarOwner classes.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
import unittest


class VarOwnerTest(unittest.TestCase):
    """
    Tests parts of :class:`myokit.VarOwner`.
    """
    def test_move_variable(self):
        # Test the method to move component variables to another component.

        # Create a model
        m = myokit.Model('LotkaVolterra')
        X = m.add_component('X')
        a = X.add_variable('a')
        a.set_rhs(3)
        b = X.add_variable('b')
        b1 = b.add_variable('b1')
        b2 = b.add_variable('b2')
        b1.set_rhs(1)
        b2.set_rhs(
            myokit.Minus(
                myokit.Minus(myokit.Name(a), myokit.Name(b1)),
                myokit.Number(1)
            )
        )
        b.set_rhs(myokit.Plus(myokit.Name(b1), myokit.Name(b2)))
        x = X.add_variable('x')
        x.promote()
        Y = m.add_component('Y')
        c = Y.add_variable('c')
        c.set_rhs(myokit.Minus(myokit.Name(a), myokit.Number(1)))
        d = Y.add_variable('d')
        d.set_rhs(2)
        y = Y.add_variable('y')
        y.promote()
        x.set_rhs(myokit.Minus(
            myokit.Multiply(myokit.Name(a), myokit.Name(x)),
            myokit.Multiply(
                myokit.Multiply(myokit.Name(b), myokit.Name(x)),
                myokit.Name(y)
            )
        ))
        x.set_initial_value(10)
        y.set_rhs(myokit.Plus(
            myokit.Multiply(
                myokit.PrefixMinus(myokit.Name(c)), myokit.Name(y)
            ),
            myokit.Multiply(
                myokit.Multiply(myokit.Name(d), myokit.Name(x)), myokit.Name(y)
            )
        ))
        y.set_initial_value(5)
        Z = m.add_component('Z')
        t = Z.add_variable('total')
        t.set_rhs(myokit.Plus(myokit.Name(x), myokit.Name(y)))
        E = m.add_component('engine')
        time = E.add_variable('time')
        time.set_rhs(0)
        time.set_binding('time')

        # Move time variable into X
        m.validate()    # If not valid, this will raise an exception
        E.move_variable(time, Z)
        m.validate()

        # Can't do it a second time
        self.assertRaises(ValueError, E.move_variable, time, Z)

        # Move to self
        Z.move_variable(time, Z)
        m.validate()
        Z.move_variable(time, Z)
        m.validate()

        # Duplicate variable name
        E.add_variable('time')
        self.assertRaises(myokit.DuplicateName, Z.move_variable, time, E)

        # Create a nested variable by moving
        m = myokit.Model()
        c = m.add_component('c')
        v = c.add_variable('v')
        w = c.add_variable('w')
        self.assertFalse(w.is_nested())
        c.move_variable(w, v)
        self.assertTrue(w.is_nested())
        v.move_variable(w, c)
        self.assertFalse(w.is_nested())

        # State variables can't be made nested
        w.promote(0)
        self.assertRaisesRegex(
            Exception, 'State variables', c.move_variable, w, v)

    def test_remove_variable(self):
        # Test the removal of a variable.

        # Create a model
        m = myokit.Model('LotkaVolterra')

        # Add a variable 'a'
        z = m.add_component('z')

        # Simplest case
        a = z.add_variable('a')
        self.assertEqual(z.count_variables(), 1)
        z.remove_variable(a)
        self.assertEqual(z.count_variables(), 0)
        self.assertRaises(Exception, z.remove_variable, a)

        # Test re-adding
        a = z.add_variable('a')
        a.set_rhs(myokit.Number(5))
        self.assertEqual(z.count_variables(), 1)

        # Test deleting dependent variables
        b = z.add_variable('b')
        self.assertEqual(z.count_variables(), 2)
        b.set_rhs(myokit.Plus(myokit.Number(3), myokit.Name(a)))

        # Test blocking of removal
        self.assertRaises(myokit.IntegrityError, z.remove_variable, a)
        self.assertEqual(z.count_variables(), 2)

        # Test removal in the right order
        z.remove_variable(b)
        self.assertEqual(z.count_variables(), 1)
        z.remove_variable(a)
        self.assertEqual(z.count_variables(), 0)

        # Test reference to current state variable values
        a = z.add_variable('a')
        a.set_rhs(myokit.Number(5))
        a.promote()
        b = z.add_variable('b')
        b.set_rhs(myokit.Plus(myokit.Number(3), myokit.Name(a)))
        self.assertRaises(myokit.IntegrityError, z.remove_variable, a)
        z.remove_variable(b)
        z.remove_variable(a)
        self.assertEqual(z.count_variables(), 0)

        # Test reference to current state variable values with "self"-ref
        a = z.add_variable('a')
        a.promote()
        a.set_rhs(myokit.Name(a))
        z.remove_variable(a)

        # Test it doesn't interfere with normal workings
        a = z.add_variable('a')
        a.promote()
        a.set_rhs(myokit.Name(a))
        b = z.add_variable('b')
        b.set_rhs(myokit.Name(a))
        self.assertRaises(myokit.IntegrityError, z.remove_variable, a)
        z.remove_variable(b)
        z.remove_variable(a)

        # Test reference to dot
        a = z.add_variable('a')
        a.set_rhs(myokit.Number(5))
        a.promote()
        b = z.add_variable('b')
        b.set_rhs(myokit.Derivative(myokit.Name(a)))
        self.assertRaises(myokit.IntegrityError, z.remove_variable, a)
        z.remove_variable(b)
        z.remove_variable(a)

        # Test if orphaned
        self.assertIsNone(b.parent())

        # Test deleting variable with nested variables
        a = z.add_variable('a')
        b = a.add_variable('b')
        b.set_rhs(myokit.Plus(myokit.Number(2), myokit.Number(2)))
        a.set_rhs(myokit.Multiply(myokit.Number(3), myokit.Name(b)))
        self.assertRaises(myokit.IntegrityError, z.remove_variable, a)
        self.assertEqual(a.count_variables(), 1)
        self.assertEqual(z.count_variables(), 1)

        # Test recursive deleting
        z.remove_variable(a, recursive=True)
        self.assertEqual(a.count_variables(), 0)
        self.assertEqual(z.count_variables(), 0)

        # Test deleting variable with nested variables that depend on each
        # other
        a = z.add_variable('a')
        b = a.add_variable('b')
        c = a.add_variable('c')
        d = a.add_variable('d')
        a.set_rhs('b + c - d')
        a.promote(0.1)
        b.set_rhs('2 * a - d')
        c.set_rhs('a + b + d')
        d.set_rhs('3 * a')
        self.assertRaises(myokit.IntegrityError, z.remove_variable, a)
        self.assertEqual(a.count_variables(), 3)
        self.assertEqual(z.count_variables(), 1)
        z.remove_variable(a, recursive=True)
        self.assertEqual(a.count_variables(), 0)
        self.assertEqual(z.count_variables(), 0)

        # Test if removed from model's label and binding lists
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        y = c.add_variable('y')
        x.set_rhs(0)
        y.set_rhs(0)
        x.set_binding('time')
        y.set_label('membrane_potential')
        self.assertIs(m.binding('time'), x)
        self.assertIs(m.label('membrane_potential'), y)
        c.remove_variable(x)
        self.assertIs(m.binding('time'), None)
        self.assertIs(m.label('membrane_potential'), y)
        c.remove_variable(y)
        self.assertIs(m.binding('time'), None)
        self.assertIs(m.label('membrane_potential'), None)

        # Test variable can't be removed if used in initial expression
        m = myokit.Model()
        c = m.add_component('c')
        p = c.add_variable('p')
        p.set_rhs(2)
        x = c.add_variable('x')
        x.set_rhs(1)
        x.promote('1 + c.p')
        self.assertRaises(myokit.IntegrityError, c.remove_variable, p)
        q = c.add_variable('q')
        q.set_rhs(10)
        x.set_initial_value('1 / c.q')
        c.remove_variable(p)
        self.assertRaises(myokit.IntegrityError, c.remove_variable, q)
        x.set_initial_value('sqrt(2)')
        c.remove_variable(q)

        # But it's OK if we're deleting a whole component
        m = myokit.Model()
        c = m.add_component('c')
        p = c.add_variable('p')
        p.set_rhs(2)
        x = c.add_variable('x')
        x.set_rhs(1)
        x.promote('1 + c.p')
        m.remove_component('c')

    def test_remove_variable_with_alias(self):
        # Test cloning of a variable with an alias after an add / remove event.

        m = myokit.Model('AddRemoveClone')
        c = m.add_component('c')
        p = c.add_variable('p')
        p.set_binding('time')
        p.set_rhs(0)
        q = c.add_variable('q')
        q.set_rhs(12)
        m.validate()    # Raises error if not ok
        m.clone()       # Raises error if not ok
        d = m.add_component('d')
        d.add_alias('bert', p)
        e = d.add_variable('e')
        e.set_rhs('10 * bert')
        m.validate()
        m.clone()
        d.add_alias('ernie', q)
        m.validate()
        m.clone()
        c.remove_variable(q)
        m.validate()
        m.clone()   # Will raise error if alias isn't deleted

    def test_resolve(self):
        # Test if an error is raised when a variable can't be resolved.

        m = myokit.Model('Resolve')
        c = m.add_component('c')
        p = c.add_variable('p')
        q = c.add_variable('q')
        p.set_rhs('10 * q')
        self.assertRaises(myokit.ParseError, q.set_rhs, '10 * r')

    def test_sequence_interface(self):
        # Test the sequence interface implementation

        model = myokit.load_model('example')
        c = model['membrane']

        vs = [v for v in c]
        self.assertEqual(vs, list(c.variables()))
        self.assertEqual(len(vs), len(c))
        v = c['V']
        self.assertEqual(v.name(), 'V')

    def test_varowner_get(self):
        # Test VarOwner.get().

        # Test basics
        m = myokit.Model('test')
        c = m.add_component('c')
        x = c.add_variable('x')
        self.assertIs(m.get('c'), c)
        self.assertIs(m.get('c.x'), x)
        self.assertIs(c.get('x'), x)

        # Test asking for object
        self.assertIs(m.get(c), c)
        self.assertIs(m.get(x), x)
        self.assertIs(c.get(x), x)
        self.assertIs(x.get(c), c)

        # Test asking for object from another model
        c2 = m.clone().get('c')
        self.assertRaisesRegex(ValueError, 'different model', c2.get, x)

        # Test not founds
        self.assertRaises(KeyError, m.get, 'y')
        self.assertRaises(KeyError, m.get, 'c.y')
        self.assertRaises(KeyError, c.get, 'y')

        # Test class filter
        self.assertIs(m.get('c', myokit.Component), c)
        self.assertRaises(KeyError, m.get, 'c', myokit.Variable)
        self.assertIs(m.get('c.x', myokit.Variable), x)
        self.assertIs(c.get('x', myokit.Variable), x)
        self.assertRaises(KeyError, m.get, 'c.x', myokit.Component)
        self.assertRaises(KeyError, c.get, 'x', myokit.Component)

    def test_add_variable_allow_renaming(self):
        # Test the ``VarProvider.add_variable_allow_renaming`` method.

        m = myokit.Model('test')
        c = m.add_component('c')
        x = c.add_variable('x')
        self.assertTrue(c.has_variable('x'))
        self.assertRaises(myokit.DuplicateName, c.add_variable, 'x')
        y = c.add_variable_allow_renaming('x')
        self.assertEqual(x.name(), 'x')
        self.assertEqual(y.name(), 'x_1')
        z = c.add_variable_allow_renaming('x')
        self.assertEqual(z.name(), 'x_2')

        # Test repeated calls
        r = c.add_variable('r')
        for i in range(10):
            r = c.add_variable_allow_renaming('r')
            self.assertEqual(r.name(), 'r_' + str(1 + i))


class ComponentTest(unittest.TestCase):
    """
    Tests parts of :class:`myokit.Component`.
    """
    def test_alias_methods(self):
        # Test various methods to do with aliases.

        # Proper use of add_alias
        m = myokit.Model()
        c1 = m.add_component('c1')
        a = c1.add_variable('a')
        a.set_rhs(2)
        c2 = m.add_component('c2')
        c2.add_alias('bert', a)
        b = c2.add_variable('b')
        b.set_rhs('3 * bert')
        self.assertEqual(b.eval(), 6)

        # Test alias() and alias_for(), remove_alias()
        self.assertTrue(c2.has_alias('bert'))
        self.assertTrue(c2.has_alias_for(a))
        self.assertEqual(c2.alias('bert'), a)
        self.assertEqual(c2.alias_for(a), 'bert')
        c2.remove_alias('bert')
        self.assertFalse(c2.has_alias('bert'))
        self.assertFalse(c2.has_alias_for(a))
        self.assertRaises(KeyError, c2.alias, 'bert')
        self.assertRaises(KeyError, c2.alias_for, a)
        c2.add_alias('bert', a)
        c2.add_alias('ernie', a)
        c2.add_alias('hello', a)
        self.assertTrue(c2.has_alias_for(a))
        c2.remove_alias('bert')
        self.assertTrue(c2.has_alias_for(a))
        c2.remove_aliases_for(a)
        self.assertFalse(c2.has_alias_for(a))

    def test_add_alias_errors(self):
        # Test error handling in :meth:`Component.add_alias()`.
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        d = m.add_component('d')
        y = d.add_variable('y')
        z = y.add_variable('z')

        # Duplicate name
        d.add_alias('great_name', x)
        self.assertRaises(myokit.DuplicateName, d.add_alias, 'y', x)

        # Alias for a nested variable
        c.add_alias('not_a_problem', y)
        self.assertRaisesRegex(
            myokit.IllegalAliasError, 'whose parent', c.add_alias, 'xyz', z)

        # Alias for a variable in the same component
        self.assertRaisesRegex(
            myokit.IllegalAliasError, 'same component', c.add_alias, 'zz', x)

        # Alias for a component
        self.assertRaisesRegex(
            myokit.IllegalAliasError, 'for variables', c.add_alias, 'zz', c)


if __name__ == '__main__':
    unittest.main()



#
# Provides SBML support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._api import (  # noqa
    Compartment,
    CSymbolVariable,
    Model,
    ModifierSpeciesReference,
    Parameter,
    Quantity,
    Reaction,
    SBMLError,
    Species,
    SpeciesReference)
from ._importer import SBMLImporter
from ._parser import (  # noqa
    SBMLParsingError,
    SBMLParser)


# Importers
_importers = {
    'sbml': SBMLImporter,
}


def importers():
    """
    Returns a dict of all importers available in this module.
    """
    return dict(_importers)

# Exporters
# Expression writers



#
# Defines the python classes that represent pacing protocols.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

import numpy as np

from bisect import bisect_right


class Protocol:
    """
    Represents a pacing protocol as a sequence of :class:`events
    <ProtocolEvent>`.

    Every event represents a time span during which the stimulus is non-zero.
    All events specify a stimulus level, a starting time and a duration. The
    stimulus level is given as a dimensionless number where 0 is no stimulus
    and 1 is a "full" stimulus. No range is specified: 2 would indicate "twice
    the normal stimulus".

    Periodic events can be created by specifying their period. A period of 0 is
    used for non-periodic events. The number of times a periodic event occurs
    can be set using the `multiplier` value. For any non-zero value, the event
    will occur exactly this many times.

    If an event starts while another event is already active, the new event
    will determine the pacing value and the old event will be deactivated.

    Scheduling two events to occur at the same time will result in an error, as
    the resulting behavior would be undefined. When events are scheduled with
    the same starting time an error will be raised immediately. If the clash
    only happens when an event re-occurs, an error will be raised when using
    the protocol (for example during a simulation).

    An event with start time ``a`` and duration ``b`` will be active during
    the interval ``[a, b)``. In other words, time ``a`` will be the first time
    it is active and time ``b`` will be the first time after ``a`` at which it
    is not.

    Protocols can be compared with ``==``, which will check if the :meth:`code`
    for both protocols is the same. Protocols can be serialized with
    ``pickle``.
    """
    def __init__(self):
        super().__init__()
        self._head = None

    def add(self, e):
        """
        (Re-)schedules an :class:`event <ProtocolEvent>`.
        """
        # No head? Then set as head
        if self._head is None:
            self._head = e
            return
        # Starts before head? Then replace as head
        if e._start < self._head._start:
            e._next = self._head
            self._head = e
            return
        # Find last event before this one
        f = self._head
        while f._next is not None and e._start >= f._next._start:
            f = f._next
        if e._start == f._start:
            raise myokit.SimultaneousProtocolEventError(
                'Two events cannot (re-)start at the same time:'
                ' Error at time t=' + str(e._start) + '.')
        e._next = f._next
        f._next = e

    def add_step(self, level, duration):
        """
        Appends an event to the end of this protocol.

        This method can be used to easily create voltage-step protocols. A call
        to ``p.add_step(level, duration)`` is equivalent to
        ``p.schedule(level, p.characteristic_time(), duration, 0, 0)``.

        Arguments:

        ``level``
            The stimulus level. 1 Represents a full-sized stimulus. Only
            non-zero levels should be set.
        ``duration``
            The length of the stimulus.

        """
        self.schedule(level, self.characteristic_time(), duration)

    def characteristic_time(self):
        """
        Returns the characteristic time associated with this protocol.

        The characteristic time is defined as the maximum characteristic time
        of all events in the protocol
        (see :meth:`ProtocolEvent.characteristic_time()`). For a sequence of
        events, this is simply the protocol duration.

        Examples:

        >>> import myokit
        >>> # A sequence of events
        >>> p = myokit.Protocol()
        >>> p.schedule(1, 0, 100)
        >>> p.schedule(1, 100, 100)
        >>> p.characteristic_time()
        200.0

        >>> # A finitely reoccurring event
        >>> p = myokit.Protocol()
        >>> p.schedule(1, 100, 0.5, 1000, 3)
        >>> p.characteristic_time()
        3100.0

        >>> # An indefinitely reoccurring event, method returns period
        >>> p = myokit.Protocol()
        >>> p.schedule(1, 100, 0.5, 1000, 0)
        >>> p.characteristic_time()
        1000.0

        """
        # Find maximum
        tmax = 0
        e = self._head
        while e is not None:
            tmax = max(tmax, e.characteristic_time())
            e = e._next
        return tmax

    def clone(self):
        """
        Returns a deep clone of this protocol.
        """
        p = Protocol()
        e = self._head
        while e is not None:
            p.add(e.clone())
            e = e._next
        return p

    def code(self):
        """
        Returns the ``mmt`` code representing this protocol and its events.
        """
        out = [
            '[[protocol]]',
            '# Level  Start    Length   Period   Multiplier',
        ]
        e = self._head
        while e is not None:
            out.append(e.code())
            e = e._next
        return '\n'.join(out)

    def create_log_for_interval(self, a, b, for_drawing=False):
        """
        Deprecated alias of :meth:`log_for_interval`.
        """
        # Deprecated since 2019-01-09
        import warnings
        warnings.warn(
            'The method `create_log_for_interval` is deprecated.'
            ' Please use `log_for_interval` instead.')
        return self.log_for_interval(a, b, for_drawing)

    def create_log_for_times(self, times):
        """
        Deprecated alias of :meth:`log_for_times`.
        """
        # Deprecated since 2019-01-09
        import warnings
        warnings.warn(
            'The method `create_log_for_times` is deprecated.'
            ' Please use `log_for_times` instead.')
        return self.log_for_times(times)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Protocol):
            return False
        return self.code() == other.code()

    def events(self):
        """
        Returns a list of all events in this protocol.
        """
        # Note: This should remain a list, not an iterator. The pacing module
        # in pacing.h depends on it!
        out = []
        e = self._head
        while e is not None:
            out.append(e)
            e = e._next
        return out

    def head(self):
        """
        Returns the first event in this protocol.

        If the protocol is empty, ``None`` will be returned.
        """
        return self._head

    def in_words(self):
        """
        Returns a description of this protocol in words.
        """
        if self._head is None:
            return 'Empty protocol.'
        out = []
        e = self._head
        while e is not None:
            out.append(e.in_words())
            e = e._next
        return '\n'.join(out)

    def is_infinite(self):
        """
        Returns True if (and only if) this protocol contains indefinitely
        recurring events.
        """
        e = self._head
        while e is not None:
            if e._period != 0 and e._multiplier == 0:
                return True
            e = e._next
        return False

    def is_sequence(self):
        """
        Checks if this protocol is a sequence of non-periodic steps.

        The following checks are performed:

        1. The protocol does not contain any periodic events
        2. The protocol does not contain any overlapping events

        See also :meth:`is_sequence_exception`.
        """
        try:
            self.is_sequence_exception()
        except NotASequenceError:
            return False
        return True

    def is_sequence_exception(self):
        """
        Like :meth:`is_sequence()`, but raises an exception if the protocol is
        not a sequence, providing some information about the check that failed.
        """
        t = 0
        e = self._head
        while e is not None:

            if e._period != 0:
                raise NotASequenceError('Protocol contains periodic event(s).')

            if e._start < t:
                raise NotASequenceError(
                    'Event starting at t=' + str(e._start)
                    + ' overlaps with previous event which finishes at t='
                    + str(t) + '.')

            t = e._start + e._duration
            e = e._next

            # Calculated position indistinguishable from user-specified next
            # even start? Then jump there instead
            if e and myokit.float.eq(t, e._start):
                t = e._start

        return True

    def is_unbroken_sequence(self):
        """
        Checks if this protocol is an unbroken sequence of steps. Returns
        ``True`` only for an unbroken sequence.

        The following checks are performed:

        1. The protocol does not contain any periodic events
        2. The protocol does not contain any overlapping events
        3. Each new event starts where the last ended

        See also :meth:`is_unbroken_sequence_exception`.
        """
        try:
            self.is_unbroken_sequence_exception()
        except NotAnUnbrokenSequenceError:
            return False
        return True

    def is_unbroken_sequence_exception(self):
        """
        Like :meth:`is_unbroken_sequence`, but raises an exception if the
        protocol is not an unbroken sequence, providing some information about
        the check that failed.
        """
        e = self._head
        if e is None:
            return True
        if e._period != 0:
            raise NotAnUnbrokenSequenceError(
                'Protocol contains periodic event(s).')

        while e._next is not None:
            t = e._start + e._duration
            e = e._next

            # Calculated position indistinguishable from user-specified next
            # even start? Then jump there instead
            if myokit.float.eq(t, e._start):
                t = e._start

            # Check for periodic events
            if e._period != 0:
                raise NotAnUnbrokenSequenceError(
                    'Protocol contains periodic event(s).')

            # Check starting time
            if e._start < t:
                raise NotAnUnbrokenSequenceError(
                    'Event starting at t=' + str(e._start)
                    + ' overlaps with previous event which finishes at t='
                    + str(t) + '.')
            elif e._start > t:
                raise NotAnUnbrokenSequenceError(
                    'Event starting at t=' + str(e._start)
                    + ' does not start directly after previous event,'
                    + ' which finishes at t=' + str(t) + '.')
        return True

    def __iter__(self):
        """
        Returns an iterator over the events in this protocol. Note that
        recurring events are only returned once.
        """
        return iter(self.events())

    def __len__(self):
        """
        Returns the number of events in this protocol. Note that recurring
        events are counted only once.
        """
        return len(self.events())

    def levels(self):
        """
        Returns the levels of the events scheduled in this protocol.

        For unbroken sequences of events this will produce a list of the levels
        visited by the protocol. For sequences with gaps or protocols with
        periodic events the relationship between actual levels and this
        method's output is more complicated.
        """
        e = self._head
        levels = []
        while e is not None:
            levels.append(e._level)
            e = e._next
        return levels

    def log_for_interval(self, a, b, for_drawing=False):
        """
        Returns a :class:`myokit.DataLog` containing the entries ``time`` and
        ``pace``, representing the value of the pacing stimulus at each  point
        on the interval ``[a, b]``.

        The time points in the log will be ``a`` and ``b``, and any time in
        between at which the pacing value changes.

        If ``for_drawing`` is set to ``True`` each time value where the
        protocol changes will be listed twice, so that a vertical line can be
        drawn from the old to the new pacing value.

        Note that the points returned are from ``a`` to ``b`` inclusive (the
        interval ``[a, b]``), and so if ``b`` coincides with the end of the
        protocol a point ``(b, 0)`` will be included in the output (protocol
        steps are defined as half-open, so include their starting point but not
        their end point).
        """
        # Test the input
        a, b = float(a), float(b)
        if b < a:
            raise ValueError('The argument `b` cannot be smaller than `a`')

        # Create a simulation log
        log = myokit.DataLog()
        log.set_time_key('time')
        log['time'] = time = []
        log['pace'] = pace = []

        # Create a pacing system
        p = PacingSystem(self)

        # Fill in points
        t = a
        v = p.advance(t)
        time.append(t)
        pace.append(v)
        while t < b:
            t = min(p.next_time(), b)
            w = p.advance(t)
            if for_drawing and v != w:
                time.append(t)
                pace.append(v)
            v = w
            time.append(t)
            pace.append(v)
        return log

    def log_for_times(self, times):
        """
        Returns a :class:`myokit.DataLog` containing the entries ``time`` and
        ``pace`` representing the value of the pacing stimulus at each point.

        The time entries ``times`` must be an non-descreasing series of
        non-negative points.
        """
        log = myokit.DataLog()
        log.set_time_key('time')
        log['time'] = times
        log['pace'] = self.value_at_times(times)
        return log

    def pop(self):
        """
        Removes and returns the event at the head of the queue.
        """
        e = self._head
        if self._head is not None:
            self._head = self._head._next
        return e

    def range(self):
        """
        Returns the minimum and maximum levels set in this protocol.
        Will return ``0, 0`` for empty protocols.
        """
        e = self._head
        if e is None:
            return 0, 0

        if e._start > 0:
            lo = hi = 0
        else:
            lo = hi = e._level

        while e is not None:
            lo = min(e._level, lo)
            hi = max(e._level, hi)
            e = e._next

        return lo, hi

    def __reduce__(self):
        """
        Pickles the Protocol.

        See: https://docs.python.org/3/library/pickle.html#object.__reduce__
        """
        return (myokit.parse_protocol, (self.code(), ))

    def schedule(self, level, start, duration, period=0, multiplier=0):
        """
        Schedules a new event.

        ``level``
            The stimulus level. 1 Represents a full-sized stimulus. Only
            non-zero levels should be set.
        ``start``
            The time this event first occurs.
        ``duration``
            The length of the stimulus.
        ``period (optional)``
            This event's period, or ``0`` if it is a one-off event.
        ``multiplier (optional)``
            For periodic events, this indicates the number of times this event
            occurs. Non-periodic events or periodic events that continue
            indefinitely can use ``0`` here.
        """
        e = ProtocolEvent(level, start, duration, period, multiplier)
        self.add(e)

    def __str__(self):
        return self.code()

    def tail(self):
        """
        Returns the last event in this protocol. Note that recurring events
        can be rescheduled, so that the event returned by this method is not
        necessarily the last event that would occur when running the protocol.

        If the protocol is empty, ``None`` will be returned.
        """
        e = self._head
        while e._next is not None:
            e = e._next
        return e

    def value_at_times(self, times):
        """
        Returns a list containing the value of the pacing variable at each time
        listed in ``times``.

        Arguments:

        ``times``
            A (non-decreasing) sequence of (non-negative) points in time.

        """
        # Times empty? Then return empty list
        if len(times) == 0:
            return []

        # Test time values are non-negative and non-decreasing
        times = np.asarray(times)
        if np.any(times[1:] < times[:-1]):
            raise ValueError(
                'The argument `times` must contain a'
                ' non-decreasing sequence of time points.')
        if times[0] < 0:
            raise ValueError('Times cannot be negative.')

        # Create a pacing system, calculate the values, and return
        p = PacingSystem(self)
        return [p.advance(t) for t in times]


class ProtocolEvent:
    """
    Describes an event occurring as part of a protocol.
    """
    def __init__(self, level, start, duration, period=0, multiplier=0):
        """
        Creates a new event

        ``level``
            The level that this event will cause the engine to hold the pacing
            variable at.
        ``start``
            The time that this event is launched.
        ``duration``
            The time this event lasts. During this time, the pacing level will
            be at the level specified by this event, unless a later event kicks
            in and overrules it.
        ``period (optional)``
            For singular events, this should be zero. To create recurring
            events, this value is used to set the interval between firings.
        ``multiplier (optional)``
            For recurring events, this value is used to indicate the number of
            times an event occurs. For singular events (with ``period=0``) this
            value should always be zero.

        """
        self._next = None
        self._level = float(level)
        self._start = float(start)
        self._duration = float(duration)
        self._period = float(period)
        self._multiplier = int(multiplier)
        if self._start < 0:
            raise myokit.ProtocolEventError(
                'An event can not start at a negative time.')
        if self._duration < 0:
            raise myokit.ProtocolEventError(
                'An event can not have a negative duration.')
        if self._period < 0:
            raise myokit.ProtocolEventError(
                'An event\'s period must be zero or greater.')
        if self._multiplier < 0:
            raise myokit.ProtocolEventError(
                'An event\'s multiplier must be zero or greater')
        if self._period == 0 and self._multiplier > 0:
            raise myokit.ProtocolEventError(
                'Non-recurring events can not specify a multiplier.')
        if float(self._multiplier) != float(multiplier):
            raise myokit.ProtocolEventError(
                'The event multiplier must be an integer.')
        if self._period > 0 and self._duration > self._period:
            raise myokit.ProtocolEventError(
                'A recurring event can not have a duration that\'s longer than'
                ' its period.')

    def characteristic_time(self):
        """
        Returns a characteristic time associated with this event.

        The time is calculated as follows:

        Singular events
            ``characteristic_time = start + duration``
        Finitely recurring events
            ``characteristic_time = start + multiplier * period``
        Indefinitely recurring events, where ``start + duration < period``
            ``characteristic_time = period``
        Indefinitely recurring events, where ``start + duration >= period``
            ``characteristic_time = start + period``

        Roughly, this means that for finite events the full duration is
        returned, while indefinitely recurring events return the time until
        the first period is completed.
        """
        if self._period == 0:
            # Singular event
            return self._start + self._duration
        elif self._multiplier > 0:
            # Finitely recurring event
            return self._start + self._period * self._multiplier
        elif self._start + self._duration < self._period:
            # Indefinitely recurring event that starts directly
            return self._period
        else:
            # Indefinitely recurring event that stars after a delay
            return self._start + self._period

    def clone(self):
        """
        Returns a clone of this event.

        Note that links to other events are not included in the copy!
        """
        return ProtocolEvent(self._level, self._start, self._duration,
                             self._period, self._multiplier)

    def code(self):
        """
        Returns a consistently formatted string representing an event.
        """
        # Level  Start    Length   Period   Multiplier
        z = 0
        x = [
            self._level,
            self._start,
            self._duration,
            self._period,
            self._multiplier,
        ]
        x = [str(x) for x in x]
        s = [8, 8, 8, 8, 0]
        for i in range(0, 5):
            n = s[i] - len(x[i]) + z
            z = 0
            if n > 0:
                x[i] += ' ' * n
            elif n < 0:
                z = n
        return ' '.join(x)

    def duration(self):
        """
        Returns this even't duration.
        """
        return self._duration

    def in_words(self):
        """
        Returns a description of this event.
        """
        out = 'Stimulus of ' + str(self._level) + ' times the normal level ' \
            'applied at t=' + str(self._start) + ', lasting ' \
            + str(self._duration)
        if self._period != 0:
            if self._multiplier > 0:
                out += ' and occurring ' + str(self._multiplier) + ' times'
            else:
                out += ' and recurring indefinitely'
            out += ' with a period of ' + str(self._period)
        return out + '.'

    def level(self):
        """
        Returns this event's pacing level.
        """
        return self._level

    def multiplier(self):
        """
        Returns the number of times this event recurs. Zero is returned for
        singular events and indefinitely recurring events.
        """
        return self._multiplier

    def next(self):
        """
        If this event is part of a :class:`myokit.Protocol`, this returns the
        next scheduled event.
        """
        return self._next

    def period(self):
        """
        Returns this event's period (or zero if the event is singular).
        """
        return self._period

    def start(self):
        """
        Returns the time this event starts.
        """
        return self._start

    def stop(self):
        """
        Returns the time this event ends, i.e. `start() + duration()`.
        """
        return self._start + self._duration

    def __str__(self):
        return self.in_words()


class PacingSystem:
    """
    This class uses a :class:`myokit.Protocol` to update the value of a
    pacing variable over time.

    A pacing system is created by passing in a protocol:

        import myokit
        p = myokit.load_protocol('example')
        s = myokit.PacingSystem(p)

    The given protocol will be cloned internally before use.

    Initially, all pacing systems are at time 0. Time can be updated (but never
    moved back!) by calling :meth:`advance(new_time)`. The current time can be
    obtained with :meth:`time()`. The value of the pacing variable is obtained
    from :meth:`pace()`. The next time the pacing variable will change can be
    obtained from :meth:`next_time()`.

    A pacing system can be used to calculate the values of the pacing variable
    at different times:

    >>> import myokit
    >>> p = myokit.load_protocol('example')
    >>> s = myokit.PacingSystem(p)
    >>> import numpy as np
    >>> time = np.linspace(0, 1000, 10001)
    >>> pace = np.array([s.advance(t) for t in time])

    """
    def __init__(self, protocol, initial_time=0):
        # The initial and current time and pacing level
        self._initial_time = initial_time  # Needed if we add a reset()
        self._time = initial_time
        self._pace = 0

        # Currently active event
        self._fire = None

        # Time the currently active event is over
        self._tdown = None

        # The next time the pacing variable changes
        self._tnext = initial_time

        # Create a copy of the protocol
        self._protocol = protocol.clone()
        #TODO: For periodic events, set an _t0, and a _i, use them to calculate
        #      the next occurence

        # Advance to initial time
        self.advance(initial_time)

    def advance(self, new_time):
        """
        Advances the time in the pacing system to ``new_time``.

        Returns the current value of the pacing variable.
        """
        # Check new_time isn't in the past
        new_time = float(new_time)
        if new_time < self._time:
            raise ValueError('New time cannot be before the current time.')

        # Set the new internal time
        self._time = new_time

        # Advance pacing system
        while myokit.float.geq(new_time, self._tnext):

            # Active event finished
            if self._fire and myokit.float.geq(self._tnext, self._tdown):
                self._fire = None
                self._pace = 0

            # New event starting
            e = self._protocol._head
            if e and myokit.float.geq(new_time, e._start):
                self._protocol.pop()
                self._fire = e
                self._tdown = e._start + e._duration
                self._pace = e._level

                # Reschedule recurring events
                if e._period > 0 and e._multiplier != 1:
                    if e._multiplier > 1:
                        e._multiplier -= 1
                    e._start += e._period
                    self._protocol.add(e)

                # Check if tdown is indistinguishable from the next event start
                # If so, then set tdown (which is always calculated) to the
                # next event start (which may be user-specified).
                x = self._protocol._head
                if x and myokit.float.eq(self._tdown, x._start):
                    self._tdown = x._start

            # Next stopping time
            self._tnext = float('inf')
            if self._fire and self._tnext > self._tdown:
                self._tnext = self._tdown
            if e and self._tnext > e._start:
                self._tnext = e._start

        return self._pace

    def next_time(self):
        """ Returns the next time the pacing system will halt at. """
        return self._tnext

    def pace(self):
        """ Returns the current value of the pacing variable. """
        return self._pace

    def time(self):
        """ Returns the current time in the pacing system. """
        return self._time


class NotASequenceError(myokit.MyokitError):
    """ Error raised exclusively by is_sequence_exception(). """
    pass


class NotAnUnbrokenSequenceError(myokit.MyokitError):
    """ Error raised exclusively by is_unbroken_sequence_exception(). """
    pass


class TimeSeriesProtocol:
    """
    Represents a pacing protocol as a sequence of time value pairs and an
    interpolation method (currently only linear interpolation is supported).

    A 1D time-series should be given as input. During the simulation, the value
    of the pacing variable will be determined by interpolating between the two
    nearest points in the series. If the simulation time is outside the bounds
    of the time-series, the first or last value in the series will be used.

    Protocols can be compared with ``==``, which will check if the sequence of
    time value pairs is the same, and the interpolation method is the same.
    Protocols can be serialized with ``pickle``.
    """

    def __init__(self, times, values, method=None):
        super().__init__()

        if len(times) != len(values):
            raise ValueError('Times and values array must have same size.')
        times_tpl, values_tpl = zip(*[
            (float(t), float(v)) for t, v in sorted(zip(times, values))
        ])
        self._times = list(times_tpl)
        self._values = list(values_tpl)

        if method is None:
            self._method = 'linear'
        else:
            self._method = str(method).lower()
            if self._method not in ('linear', ):
                raise ValueError(
                    'Unknown interpolation method: ' + self._method)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, TimeSeriesProtocol):
            return False
        return (
            self._method == other._method
            and self._values == other._values
            and self._times == other._times
        )

    def __getstate__(self):
        return {
            'times': self._times,
            'values': self._values,
            'method': self._method,
        }

    def __setstate__(self, values):
        self._times = values['times']
        self._values = values['values']
        self._method = values['method']

    def clone(self):
        """ Returns a clone of this protocol. """
        return TimeSeriesProtocol(self._times, self._values, self._method)

    def pace(self, t):
        """ Returns the value of the pacing variable at time ``t``. """
        if t < self._times[0]:
            return self._values[0]
        if t > self._times[-1]:
            return self._values[-1]
        i = bisect_right(self._times, t) - 1
        if i == len(self._times) - 1:
            return self._values[i]
        return self._values[i] + (t - self._times[i]) * (
            self._values[i + 1] - self._values[i]
        ) / (self._times[i + 1] - self._times[i])

    def times(self):
        """ Returns a list of the times in this protocol. """
        return self._times

    def values(self):
        """ Returns a list of the values in this protocol. """
        return self._values



#!/usr/bin/env python3
#
# Tests the CVODES simulation class, when using a precompiled simulation.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import (
    DIR_DATA,
)


class PrecompiledSimulationTest(unittest.TestCase):
    """
    Tests the CVODES simulation class, when using a precompiled simulation.
    """

    def test(self):
        # Create and run two simulations and compare the output

        # Load model and protocol
        model, protocol, _ = myokit.load(os.path.join(DIR_DATA, 'lr-1991.mmt'))

        # Select sensitivities to calculate
        sens = (
            ('membrane.V', 'ica.Ca_i', 'ina.INa', 'ik.x', 'dot(ina.m)'),
            ('init(ik.x)', 'init(ica.Ca_i)', 'ina.gNa', 'ik.PNa_K')
        )

        # Create both simulation
        path = 'precompiled-test.zip'
        assert not os.path.exists(path), 'Path used in test already exists'
        try:
            sim1 = myokit.Simulation(model, protocol, sens, path=path)
            sim2 = myokit.Simulation.from_path(path)
        finally:
            if os.path.exists(path):
                os.remove(path)
        assert not os.path.exists(path), 'Test leaked file: ' + path

        sim1.pre(10)
        sim1.run(10)
        d1, s1 = sim1.run(1000)
        d1, s1 = d1.npview(), np.array(s1)

        sim2.pre(10)
        sim2.run(10)
        d2, s2 = sim2.run(1000)
        d2, s2 = d2.npview(), np.array(s2)

        for k, v in d1.items():
            if k == 'engine.realtime':
                continue
            if np.max(np.abs(v - d2[k])) > 0:
                print(k)
            self.assertTrue(np.all(v == d2[k]))

        self.assertTrue(np.all(s1 == s2))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Tests the UserFunction class.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest
import myokit

# Further testing in test_model.py


class UserFunctionTest(unittest.TestCase):
    """ Tests :class:`UserFunction`. """

    def test_user_function(self):
        # Test :class:`UserFunction` creation and methods.

        # Create without arguments
        f = myokit.UserFunction('bert', [], myokit.Number(12))
        args = list(f.arguments())
        self.assertEqual(len(args), 0)
        self.assertEqual(f.convert([]), myokit.Number(12))

        # Create with one argument
        f = myokit.UserFunction(
            'x', [myokit.Name('a')], myokit.parse_expression('1 + a'))
        self.assertEqual(len(list(f.arguments())), 1)
        args = {myokit.Name('a'): myokit.Number(3)}
        self.assertEqual(f.convert(args).eval(), 4)

        # Create with two argument
        f = myokit.UserFunction(
            'x', [myokit.Name('a'), myokit.Name('b')],
            myokit.parse_expression('a + b'))
        self.assertEqual(len(list(f.arguments())), 2)
        args = {
            myokit.Name('a'): myokit.Number(3),
            myokit.Name('b'): myokit.Number(4)
        }
        self.assertEqual(f.convert(args), myokit.parse_expression('3 + 4'))

        # Call with wrong arguments
        del args[myokit.Name('a')]
        self.assertRaisesRegex(
            ValueError, 'Wrong number', f.convert, args)
        args[myokit.Name('c')] = myokit.Number(100)
        self.assertRaisesRegex(
            ValueError, 'Missing input argument', f.convert, args)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests Myokit's SBML support.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit
import myokit.formats
import myokit.formats.sbml

# from shared import DIR_FORMATS, WarningCollector
from myokit.tests import DIR_FORMATS, WarningCollector


class SBMLImporterTest(unittest.TestCase):
    """
    Tests the SBMLImporter.
    """

    def test_capability_reporting(self):
        # Test if the right capabilities are reported.
        i = myokit.formats.importer('sbml')
        self.assertFalse(i.supports_component())
        self.assertTrue(i.supports_model())
        self.assertFalse(i.supports_protocol())

    def test_model(self):
        # Tests importing the Hodgkin-Huxley model
        i = myokit.formats.importer('sbml')
        with WarningCollector() as w:
            model = i.model(
                os.path.join(DIR_FORMATS, 'sbml', 'HodgkinHuxley.xml'))
        self.assertIn('Unknown SBML namespace', w.text())

        self.assertIsInstance(model, myokit.Model)


class SBMLTestSuiteExamplesTest(unittest.TestCase):
    """
    Tests whether forward simulation of test cases from the SBML test suite
    coincides with the provided time-series data.

    Test Suite examples were taken from http://sbml.org/Facilities/Database/.
    """

    @classmethod
    def setUpClass(cls):
        cls.importer = myokit.formats.importer('sbml')

    def get_sbml_file(self, case):
        return os.path.join(
            DIR_FORMATS, 'sbml', 'model', case + '-sbml-l3v2.xml')

    def get_results(self, case):
        path = os.path.join(
            DIR_FORMATS, 'sbml', 'result', case + '-results.csv')
        data = np.genfromtxt(path, delimiter=',')

        # Eliminate title row
        data = data[1:, :]

        return data

    def test_case_00001(self):
        #
        # From the SBML Test Suite settings:
        #
        # start: 0
        # duration: 5
        # steps: 50
        # variables: S1, S2
        # absolute: 1.000000e-007
        # relative: 0.0001
        # amount: S1, S2
        # concentration:

        case = '00001'

        model = self.importer.model(self.get_sbml_file(case))

        results = self.get_results(case)
        times = results[:, 0]
        s1 = results[:, 1]
        s2 = results[:, 2]

        sim = myokit.Simulation(model)
        output = sim.run(
            duration=times[-1] + 1,
            log=['compartment.S1_amount', 'compartment.S2_amount'],
            log_times=times)

        s1_sim = np.array(output['compartment.S1_amount'])
        np.testing.assert_almost_equal(s1_sim, s1, decimal=6)

        s2_sim = np.array(output['compartment.S2_amount'])
        np.testing.assert_almost_equal(s2_sim, s2, decimal=6)

    def test_case_00004(self):
        #
        # From the SBML Test Suite settings:
        #
        # start: 0
        # duration: 10.0
        # steps: 50
        # variables: S1, S2
        # absolute: 1.000000e-004
        # relative: 0.0001
        # amount: S1, S2
        # concentration:

        case = '00004'

        model = self.importer.model(self.get_sbml_file(case))

        results = self.get_results(case)
        times = results[:, 0]
        s1 = results[:, 1]
        s2 = results[:, 2]

        sim = myokit.Simulation(model)
        output = sim.run(
            duration=times[-1] + 1,
            log=['compartment.S1_amount', 'compartment.S2_amount'],
            log_times=times)

        s1_sim = np.array(output['compartment.S1_amount'])
        np.testing.assert_almost_equal(s1_sim, s1, decimal=4)

        s2_sim = np.array(output['compartment.S2_amount'])
        np.testing.assert_almost_equal(s2_sim, s2, decimal=4)

    def test_case_01103(self):
        #
        # From the SBML Test Suite settings:
        #
        # start: 0
        # duration: 10
        # steps: 100
        # variables: X
        # absolute: 0.0001
        # relative: 0.0001
        # amount: X
        # concentration:

        case = '01103'

        model = self.importer.model(self.get_sbml_file(case))

        results = self.get_results(case)
        times = results[:, 0]
        x = results[:, 1]

        sim = myokit.Simulation(model)
        output = sim.run(
            duration=times[-1] + 1,
            log=['default_compartment.X_amount'],
            log_times=times)

        x_sim = np.array(output['default_compartment.X_amount'])
        np.testing.assert_almost_equal(x_sim, x, decimal=4)


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#
# Provides an export to a CUDA kernel
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import CudaKernelExporter, CudaKernelRLExporter
from ._ewriter import CudaExpressionWriter
from myokit.formats import ansic


# Importers
# Exporters
_exporters = {
    'cuda-kernel': CudaKernelExporter,
    'cuda-kernel-rl': CudaKernelRLExporter,
}


def exporters():
    """
    Returns a list of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'cuda': CudaExpressionWriter,
}


def ewriters():
    """
    Returns a list of all expression writers available in this module.
    """
    return dict(_ewriters)


# Keywords
keywords = list(ansic.keywords)
# TODO: Append CUDA keywords



#
# CUDA expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.python import PythonExpressionWriter


class CudaExpressionWriter(PythonExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to their CUDA equivalents.
    """
    def __init__(self, precision=myokit.SINGLE_PRECISION):
        super().__init__()
        self._function_prefix = ''
        self._sp = (precision == myokit.SINGLE_PRECISION)

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):

    def _ex_number(self, e):
        return myokit.float.str(e) + 'f' if self._sp else myokit.float.str(e)

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):
    #def _ex_minus(self, e):
    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    def _ex_quotient(self, e):
        # Note that this _must_ round towards minus infinity.
        # See myokit.Quotient
        # CUDA docs are unclear on convention, so assuming it follows C and
        # so we need a custom implementation.
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        # Note that this _must_ use the same round-to-neg-inf convention as
        # myokit.Quotient.
        # CUDA docs are unclear on convention, so assuming it follows C and
        # so we need a custom implementation.
        return self.ex(myokit.Minus(
            e[0], myokit.Multiply(e[1], myokit.Quotient(e[0], e[1]))))

    def _ex_power(self, e):
        if e[1] == myokit.Number(2):
            if e.bracket(e[0]):
                return '((' + self.ex(e[0]) + ') * (' + self.ex(e[0]) + '))'
            else:
                return '(' + self.ex(e[0]) + ' * ' + self.ex(e[0]) + ')'
        else:
            f = 'powf' if self._sp else 'pow'
            return f + '(' + self.ex(e[0]) + ', ' + self.ex(e[1]) + ')'

    def _ex_sqrt(self, e):
        f = 'sqrtf' if self._sp else 'sqrt'
        return self._ex_function(e, f)

    def _ex_sin(self, e):
        f = 'sinf' if self._sp else 'sin'
        return self._ex_function(e, f)

    def _ex_cos(self, e):
        f = 'cosf' if self._sp else 'cos'
        return self._ex_function(e, f)

    def _ex_tan(self, e):
        f = 'tanf' if self._sp else 'tan'
        return self._ex_function(e, f)

    def _ex_asin(self, e):
        f = 'asinf' if self._sp else 'asin'
        return self._ex_function(e, f)

    def _ex_acos(self, e):
        f = 'acosf' if self._sp else 'acos'
        return self._ex_function(e, f)

    def _ex_atan(self, e):
        f = 'atanf' if self._sp else 'atan'
        return self._ex_function(e, f)

    def _ex_exp(self, e):
        f = 'expf' if self._sp else 'exp'
        return self._ex_function(e, f)

    def _ex_log(self, e):
        f = 'logf' if self._sp else 'log'
        if len(e) == 1:
            return self._ex_function(e, f)
        return '(' + f + '(' + self.ex(e[0]) + ') / ' + f + '(' \
            + self.ex(e[1]) + '))'

    def _ex_log10(self, e):
        f = 'log10f' if self._sp else 'log10'
        return self._ex_function(e, f)

    def _ex_floor(self, e):
        f = 'floorf' if self._sp else 'floor'
        return self._ex_function(e, f)

    def _ex_ceil(self, e):
        f = 'ceilf' if self._sp else 'ceil'
        return self._ex_function(e, f)

    def _ex_abs(self, e):
        f = 'fabsf' if self._sp else 'fabs'
        return self._ex_function(e, f)

    def _ex_not(self, e):
        return '!(' + self.ex(e[0]) + ')'

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_condition(e, '&&')

    def _ex_or(self, e):
        return self._ex_infix_condition(e, '||')

    def _ex_if(self, e):
        return '(%s ? %s : %s)' % (self.ex(e._i), self.ex(e._t), self.ex(e._e))

    def _ex_piecewise(self, e):
        s = []
        n = len(e._i)
        for i in range(0, n):
            s.append('(')
            s.append(self.ex(e._i[i]))
            s.append(' ? ')
            s.append(self.ex(e._e[i]))
            s.append(' : ')
        s.append(self.ex(e._e[n]))
        s.append(')' * n)
        return ''.join(s)




#
# Cable simulation using a simple forward Euler implementation
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform

import myokit

# Location of source file
SOURCE_FILE = 'cable.c'


class Simulation1d(myokit.CModule):
    """
    Can run 1d cable simulations based on a :class:`model <Model>`.

    ``model``
        The model to simulate with. This model will be cloned when the
        simulation is created so that no changes to the given model will be
        made.
    ``protocol``
        An optional pacing protocol, used to stimulate a number of cells at the
        start of the cable.
    ``ncells``
        The number of cells in the cable
    ``rl``
        Use Rush-Larsen updates instead of forward-Euler for any Hodgkin-Huxley
        gating variables (default=False).

    This simulation provides the following inputs variables can bind to:

    ``time``
        The simulation time
    ``pace``
        The pacing level, this is set if a protocol was passed in. Will be set
        to 0 if no protocol is provided.
    ``diffusion_current``
        The current flowing from the cell to its neighbors. This will be
        positive when the cell is acting as a source, negative when it is
        acting as a sink. Will be set to 0 if no connections are made.

    The variable ``time`` is set globally, meaning each cell uses the same
    value. The variables ``pace`` and ``diffusion_current`` have different
    values per cell. The number of paced cell can be set with
    :meth:`set_paced_cells`.

    A single labeled variable is required for this simulation to work:

    ``membrane_potential``
        The variable representing the membrane potential.

    Simulations maintain an internal state consisting of

    - the current simulation time
    - the current state
    - the default state

    When a simulation is created, the simulation time is set to 0 and both the
    current and the default state are equal to the state of the given model,
    copied once for each cell.
    After each call to :meth:`run` the time variable and current state are
    updated, so that each successive call to run continues where the previous
    simulation left off. A :meth:`reset` method is provided that will set the
    time back to 0 and revert the current state to the default state. To change
    the time or state manually, use :meth:`set_time` and :meth:`set_state`.

    A pre-pacing method :meth:`pre` is provided that doesn't affect the
    simulation time but will update the current *and the default state*. This
    allows you to pre-pace, run a simulation, reset to the pre-paced state, run
    another simulation etc.

    The ``diffusion_current`` is calculated as::

        i = sum[g * (V - V_j)]

    Where the sum is taken over all neighboring cells j (see [1]).

    The resulting ODE system is solved using a forward Euler (FE) method with
    fixed step sizes. Smaller step sizes lead to more accurate results, and it
    is recommended any important results are double-checked by re-running with
    a reduced step size. Any states written in a Hodgkin-Huxley form can be
    updated using Rush-Larsen steps (see [2]), by setting ``rl=True``. This
    often increases stability (allowing for larger step sizes) but can reduce
    accuracy (see [3]) so that care must be taken when using this method.

    [1] Myokit: A simple interface to cardiac cellular electrophysiology.
    Clerx, Collins, de Lange, Volders (2016) Progress in Biophysics and
    Molecular Biology.

    [2] A practical algorithm for solving dynamic membrane equations.
    Rush, Larsen (1978) IEEE Transactions on Biomedical Engineering

    [3] Cellular cardiac electrophysiology modelling with Chaste and CellML
    Cooper, Spiteri, Mirams (2015) Frontiers in Physiology

    """
    _index = 0      # Unique id for generated module

    def __init__(self, model, protocol=None, ncells=50, rl=False):
        super().__init__()

        # Require a valid model
        model.validate()

        # Set protocol
        self.set_protocol(protocol)

        # Set number of cells
        ncells = int(ncells)
        if ncells < 1:
            raise ValueError('The number of cells must be at least 1.')
        self._ncells = ncells

        # Set rush-larsen mode
        self._rl = bool(rl)

        # Get membrane potential variable
        vm = model.label('membrane_potential')
        if vm is None:
            raise ValueError(
                'This simulation requires the membrane potential'
                ' variable to be labelled as "membrane_potential".')

        # Prepare for Rush-Larsen updates, and/or clone model
        rl_states = {}
        if self._rl:
            import myokit.lib.hh as hh

            # Convert alpha-beta formulations to inf-tau forms, cloning model
            self._model = hh.convert_hh_states_to_inf_tau_form(model, vm)
            self._vm = self._model.get(vm.qname())
            del model, vm

            # Get (inf, tau) tuple for every Rush-Larsen state
            for state in self._model.states():
                res = hh.get_inf_and_tau(state, self._vm)
                if res is not None:
                    rl_states[state] = res

        else:
            # Clone model, store
            self._model = model.clone()
            self._vm = self._model.get(vm.qname())
            del model, vm

        # Set number of cells paced
        self.set_paced_cells()

        # Set conductance
        self.set_conductance()

        # Set step size
        self.set_step_size()

        # Set remaining properties
        self._time = 0
        self._nstate = self._model.count_states()

        # Get membrane potential variable
        self._vm = self._model.label('membrane_potential')
        # Already checked this above
        # if self._vm is None:
        #    raise ValueError(
        #        'This simulation requires the membrane potential'
        #        ' variable to be labelled as "membrane_potential".')

        # Check for binding to diffusion_current
        if self._model.binding('diffusion_current') is None:
            raise ValueError(
                'This simulation requires a variable to be bound to'
                ' "diffusion_current" to pass current from one cell to the'
                ' next')

        # Set state and default state
        self._state = self._model.initial_values(True) * ncells
        self._default_state = list(self._state)

        # Unique simulation id
        Simulation1d._index += 1
        module_name = 'myokit_sim1d_' + str(Simulation1d._index)
        module_name += '_' + str(myokit.pid_hash())

        # Arguments
        args = {
            'module_name': module_name,
            'model': self._model,
            'vmvar': self._vm,
            'ncells': self._ncells,
            'rl_states': rl_states,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = []
        if platform.system() != 'Windows':  # pragma: no windows cover
            libs.append('m')

        # Create simulation
        libd = None
        incd = [myokit.DIR_CFUNC]
        self._sim = self._compile(module_name, fname, args, libs, libd, incd)

    def conductance(self):
        """
        Returns the current conductance.
        """
        return self._conductance

    def default_state(self, icell=None):
        """
        Returns the default simulation state as a list of ``len(state) *
        ncells`` floating point values. If the optional argument ``icell`` is
        set to a valid cell index only the state of that cell is returned.
        """
        if icell is None:
            return list(self._default_state)
        else:
            icell = int(icell)
            if icell < 0 or icell >= self._ncells:
                raise ValueError('Given cell index out of range.')
            offset = icell * self._nstate
            return self._default_state[offset:offset + self._nstate]

    def paced_cells(self):
        """
        Returns the number of cells that will receive a stimulus from the
        pacing protocol.
        """
        return self._npaced

    def pre(self, duration, progress=None, msg='Pre-pacing Simulation1d'):
        """
        This method can be used to perform an unlogged simulation, typically to
        pre-pace to a (semi-)stable orbit.

        After running this method

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        Calls to :meth:`reset` after using :meth:`pre` will revert the
        simulation to this new default state.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        self._run(duration, myokit.LOG_NONE, 1, progress, msg)
        self._default_state = list(self._state)

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to 0
        - The state is set to the default state

        """
        self._time = 0
        self._state = list(self._default_state)

    def run(
            self, duration, log=None, log_interval=1.0, progress=None,
            msg='Running Simulation1d'):
        """
        Runs a simulation and returns the logged results. Running a simulation
        has the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
          elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.

        The variables to log can be indicated using the ``log`` argument. There
        are several options for its value:

        - ``None`` (default), to log all states
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_INTER``,
          ``myokit.LOG_BOUND``, ``myokit.LOG_ALL``.
        - A list of qnames or variable objects
        - A :class:`myokit.DataLog` object.

        For more details on the ``log`` argument, see the function
        :meth:`myokit.prepare_log`.

        Any variables bound to "time" will be logged globally, all others will
        be logged per cell. These variables will be prefixed with a number
        indicating the cell index. For example, when using::

            s = Simulation1d(m, p, ncells=3)
            d = s.run(1000, log=['engine.time', 'membrane.V']

        where <engine.time> is bound to "time" and <membrane.V> is the membrane
        potential variable, the resulting log will contain the following
        variables::

            {
                'engine.time'  : [...],
                '0.membrane.V' : [...],
                '1.membrane.V' : [...],
                '2.membrane.V' : [...],
            }

        A log entry is created every time *at least* ``log_interval`` time
        units have passed.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        r = self._run(duration, log, log_interval, progress, msg)
        self._time += duration
        return r

    def _run(self, duration, log, log_interval, progress, msg):
        # Simulation times
        if duration < 0:
            raise ValueError('Simulation time can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Gather global variables in model
        global_vars = [self._model.time().qname()]
        pace = self._model.binding('pace')
        if pace is not None:
            global_vars.append(pace.qname())

        # Parse log argument
        allowed = myokit.LOG_STATE + myokit.LOG_BOUND + myokit.LOG_INTER
        log = myokit.prepare_log(
            log,
            self._model,
            dims=(self._ncells,),
            global_vars=global_vars,
            if_empty=myokit.LOG_STATE + myokit.LOG_BOUND,
            allowed_classes=allowed,
        )

        # Get event tuples
        # Logging period
        log_interval = 0 if log_interval is None else float(log_interval)

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument "progress" must be either a subclass of'
                    ' myokit.ProgressReporter or None.')

        # Run simulation
        if duration > 0:
            # Initialize
            state_in = self._state
            state_out = list(state_in)
            self._sim.sim_init(
                self._ncells,
                self._conductance,
                tmin,
                tmax,
                self._step_size,
                state_in,
                state_out,
                self._protocol,
                min(self._npaced, self._ncells),
                log,
                log_interval)
            t = tmin
            try:
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1.0 / duration if duration != 0 else 1
                        while t < tmax:
                            t = self._sim.sim_step()
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    while t < tmax:
                        t = self._sim.sim_step()
            finally:
                # Clean even after keyboardinterrupt or exception
                self._sim.sim_clean()

            # Update state
            self._state = state_out

        # Return log
        return log

    def _set_state(self, state, icell, update):
        """
        Handles set_state and set_default_state.
        """
        n = len(state)
        if n == self._nstate:
            if icell is None:
                # State might not be a list, at this point...
                return list(state) * self._ncells
            else:
                icell = int(icell)
                if icell < 0 or icell >= self._ncells:
                    raise ValueError('Given cell index out of range.')
                offset = icell * self._nstate
                update[offset:offset + self._nstate] = state
                return update
        elif n == self._nstate * self._ncells:
            return list(state)

    def set_conductance(self, g=10):
        """
        Changes the cell-to-cell conductance.
        """
        g = float(g)
        if g < 0:
            raise ValueError('Conductance cannot be negative.')
        self._conductance = g

    def set_default_state(self, state, icell=None):
        """
        Changes this simulation's default state.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``n_states`` and
           ``i_cell=None`` the given state will be set as the new state of all
           cells in the simulation.
        2. Called with an argument ``state`` of size n_states and
           ``i_cell`` equal to a valid cell index, this method will update only
           the selected cell's state.
        3. Finally, when called with a ``state`` of size ``n_states * ncells``
           the method will treat ``state`` as a concatenation of state vectors
           for each cell.

        """
        self._default_state = self._set_state(
            state, icell, self._default_state)

    def set_paced_cells(self, n=5):
        """
        Sets the number of cells that will receive a stimulus from the pacing
        protocol.
        """
        n = int(n)
        if n < 0:
            raise ValueError(
                'The number of cells to stimulate cannot be negative.')
        self._npaced = n

    def set_state(self, state, icell=None):
        """
        Changes the state of this simulation's model.

        This can be used in three different ways:

        1. When called with an argument ``state`` of size ``nstates`` and
           ``icell=None`` the given state will be set as the new state of all
           cells in the simulation.
        2. Called with an argument ``state`` of size ``nstates`` and
           ``icell`` equal to a valid cell index, this method will update only
           the selected cell's state.
        3. Finally, when called with a ``state`` of size ``nstates * ncells``
           the method will treat ``state`` as a concatenation of state vectors
           for each cell.

        """
        self._state = self._set_state(state, icell, self._state)

    def set_step_size(self, step_size=0.005):
        """
        Sets the solver step size. In some cases, the solver will take a
        slightly smaller step size, either to arrive exactly at the start/end
        of a pacing event or to arrive exactly at the end of a simulation.
        """
        step_size = float(step_size)
        if step_size <= 0:
            raise ValueError('Step size must be greater than zero.')
        self._step_size = step_size

    def set_protocol(self, protocol=None):
        """
        Changes the pacing protocol used by this simulation.
        """
        if protocol is None:
            self._protocol = None
        else:
            self._protocol = protocol.clone()

    def set_time(self, time=0):
        """
        Sets the current simulation time.
        """
        self._time = float(time)

    def state(self, icell=None):
        """
        Returns the current simulation state as a list of ``len(state) *
        ncells`` floating point values. If the optional argument ``icell`` is
        set to a valid cell index only the state of that cell is returned.
        """
        if icell is None:
            return list(self._state)
        else:
            icell = int(icell)
            if icell < 0 or icell >= self._ncells:
                raise ValueError('Given cell index out of range.')
            offset = icell * self._nstate
            return self._state[offset:offset + self._nstate]

    def time(self):
        """
        Returns the current simulation time.
        """
        return self._time



#
# Exports to a generic XML format.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

from lxml import etree

import myokit


class XMLExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates an XML file
    containing a model's equations, encoded in Content MathML.
    """

    def model(self, path, model, protocol=None):
        """
        Export the model to a generic xml document.
        """
        path = os.path.abspath(os.path.expanduser(path))

        # Create model xml element
        root = etree.Element('math')
        root.attrib['xmlns'] = 'http://www.w3.org/1998/Math/MathML'

        # Create expression writer
        import myokit.formats.mathml
        writer = myokit.formats.mathml.MathMLExpressionWriter()
        writer.set_mode(presentation=False)
        writer.set_time_variable(model.time())

        # Write equations
        for var in model.variables(deep=True):
            writer.eq(var.eq(), root)

        # Write xml to file
        doc = etree.ElementTree(root)
        doc.write(path, encoding='utf-8', method='xml')

        # Pretty output
        if True:
            import xml.dom.minidom as m
            xml = m.parse(path)
            with open(path, 'wb') as f:
                f.write(xml.toprettyxml(encoding='utf-8'))

    def supports_model(self):
        """
        Returns ``True``.
        """
        return True




#
# CellML expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from lxml import etree

import myokit
import myokit.formats.cellml

from myokit.formats.mathml import MathMLExpressionWriter


class CellMLExpressionWriter(MathMLExpressionWriter):
    """
    Writes equations for variables using CellML's version of MathML.

    To use, create an expression writer, then pass in a method to set good
    variable names with :meth:`set_lhs_function()`, and a method to look up
    unit names with :meth:`set_unit_function()`.
    """
    def __init__(self, version='1.0'):
        super().__init__()
        super().set_mode(presentation=False)

        # Units lookup method
        self._funits = None

        # Use unames by default
        self._flhs = lambda x: x.var().uname()

        # Get namespace based on version
        if version == '1.0':
            self._ns = myokit.formats.cellml.NS_CELLML_1_0
        elif version == '1.1':
            self._ns = myokit.formats.cellml.NS_CELLML_1_1
        elif version == '2.0':
            self._ns = myokit.formats.cellml.NS_CELLML_2_0
        else:
            raise ValueError('Unknown CellML version: ' + str(version))

        # Namespaces for element creation
        self._nsmap['cellml'] = self._ns

    def set_unit_function(self, f):
        """
        Sets a unit lookup function, which will be used to convert a
        :class:`myokit.Unit` to a CellML units name.
        """
        self._funits = f

    def _ex_number(self, e, t):
        x = super()._ex_number(e, t)
        unit = e.unit()
        if self._funits is not None and unit is not None:
            units = self._funits(e.unit())
        else:
            units = 'dimensionless'
        x.attrib[etree.QName(self._ns, 'units')] = units

    def _ex_quotient(self, e, t):
        # CellML 1.0 subset doesn't contain quotient
        # Note that this _must_ round towards minus infinity!
        # See myokit.Quotient !
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])), t)

    def _ex_remainder(self, e, t):
        # CellML 1.0 subset doesn't contain remainder
        # Note that this _must_ use the same round-to-neg-inf convention as
        # myokit.Quotient! Implementation below is consistent with Python
        # convention:
        return self.ex(myokit.Minus(
            e[0], myokit.Multiply(e[1], myokit.Quotient(e[0], e[1]))), t)

    def set_mode(self, presentation=False):
        """
        This expression writer only supports content MathML.
        """
        if presentation:
            raise RuntimeError(
                'Presentation MathML is not supported in CellML.')




#
# Contains classes for progress reporting during long operations.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import timeit

import myokit


class ProgressReporter:
    """
    Interface for progress updates in Simulations. Also allows some job types
    to be cancelled by the user.

    Many simulation types take an argument ``progress`` that can be used to
    pass in an object implementing this interface. The simulation will use this
    object to report on its progress.

    Note that progress reporters should be re-usable, but the behavior when
    making calls to a reporter from two different processes (either through
    multi-threading/multi-processing or jobs nested within jobs) is undefined.

    An optional description of the job to run can be passed in at construction
    time as ``msg``.
    """
    def enter(self, msg=None):
        """
        This method will be called when the job that provides progress updates
        is started.

        An optional description of the job to run can be passed in at
        construction time as ``msg``.
        """
        pass

    def exit(self):
        """
        Called when a job is finished and the progress reports should stop.
        """
        pass

    def job(self, msg=None):
        """
        Returns a context manager that will enter and exit this
        :class:`ProgressReporter` using the ``with`` statement.
        """
        return ProgressReporter._Job(self, msg)

    def update(self, progress):
        """
        This method will be called to provides updates about the current
        progress. This is indicated using the floating point value
        ``progress``, which will have a value in the range ``[0, 1]``.

        The return value of this update can be used to signal a job should be
        cancelled: Return ``True`` to keep going, ``False`` to cancel.
        """
        pass

    class _Job:
        def __init__(self, parent, msg):
            self._parent = parent
            self._msg = msg

        def __enter__(self):
            self._parent.enter(self._msg)

        def __exit__(self, type, value, traceback):
            self._parent.exit()


class ProgressPrinter(ProgressReporter):
    """
    Writes progress information to stdout, can be used during a simulation.

    For example::

        m, p, x = myokit.load('example')
        s = myokit.Simulation(m, p)
        w = myokit.ProgressPrinter(digits=1)
        d = s.run(10000, progress=w)

    This will print strings like::

        [8.9 minutes] 71.7 % done, estimated 4.2 minutes remaining

    To ``stdout`` during the simulation.

    Output is only written if the new percentage done differs from the old one,
    in a string format specified by the number of ``digits`` to display. The
    ``digits`` parameter accepts the special value ``-1`` to only print out a
    status every ten percent.
    """
    def __init__(self, digits=1):
        super().__init__()
        self._b = myokit.tools.Benchmarker()
        self._f = None
        self._d = int(digits)

    def enter(self, msg=None):
        # Reset
        self._b.reset()
        self._f = None

    def update(self, f):
        """
        See: :meth:`ProgressReporter.update()`.
        """
        if self._d < 0:
            f = 10 * int(10 * f)
        else:
            f = round(100 * f, self._d)
        if f != self._f:
            self._f = f
            t = self._b.time()
            if f > 0:
                p = t * (100 / f - 1)
                if p > 60:
                    p = str(round(p / 60, 1))
                    p = ', estimated ' + p + ' minutes remaining'
                else:
                    p = str(int(round(p, 0)))
                    p = ', estimated ' + p + ' seconds remaining'
            else:
                p = ''
            t = str(round(t / 60, 1))
            print('[' + t + ' minutes] ' + str(f) + ' % done' + p)
        return True


class Timeout(ProgressReporter):
    """
    Progress reporter that cancels a simulation after ``timeout`` seconds.
    """
    def __init__(self, timeout):
        super().__init__()
        self._timeout = float(timeout)

    def enter(self, msg=None):
        self._max_time = timeit.default_timer() + self._timeout

    def update(self, progress):
        return timeit.default_timer() < self._max_time




#
# This hidden module contains the GUI elements used throughout Myokit.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform
import signal
import sys

import myokit

# Detect platform
platform = platform.system()

# Select Qt library to use
pyqt5 = False
pyqt6 = False
pyside2 = False
pyside6 = False

# Allow overriding automatic selection
if myokit.FORCE_PYQT6:
    pyqt6 = True
if myokit.FORCE_PYQT5:
    pyqt5 = True
elif myokit.FORCE_PYSIDE6:
    pyside6 = True
elif myokit.FORCE_PYSIDE2:
    pyside2 = True
else:
    # Automatic selection
    try:
        import PyQt6  # noqa
        pyqt6 = True
    except ImportError:
        try:
            import PySide6  # noqa
            pyside6 = True
        except ImportError:
            try:
                import PyQt5  # noqa
                pyqt5 = True
            except ImportError:
                try:
                    import PySide2  # noqa
                    pyside2 = True
                except ImportError:
                    raise ImportError(
                        'Unable to find PyQt6, PyQt5, PySide6, or PySide2.')


# Import and configure Qt
if pyqt6:
    from PyQt6 import QtGui, QtWidgets, QtCore
    from PyQt6.QtCore import Qt

    # Use PySide signal names
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    QtCore.Property = QtCore.pyqtProperty

    # Set backend variables
    backend = 'PyQt6'
    qtversion = 6

elif pyqt5:
    from PyQt5 import QtGui, QtWidgets, QtCore
    from PyQt5.QtCore import Qt

    # Mimic PyQt6 API changes
    # https://doc.qt.io/qt-6/widgets-changes-qt6.html
    QtGui.QAction = QtWidgets.QAction
    QtWidgets.QApplication.exec = QtWidgets.QApplication.exec_

    # Use PySide signal names
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    QtCore.Property = QtCore.pyqtProperty

    # Set backend variables
    backend = 'PyQt5'
    qtversion = 5

elif pyside6:
    from PySide6 import QtGui, QtWidgets, QtCore
    from PySide6.QtCore import Qt

    # Mimic PyQt6 API changes
    QtWidgets.QApplication.exec = QtWidgets.QApplication.exec_

    # Set backend variables
    backend = 'PySide6'
    qtversion = 6

elif pyside2:
    from PySide2 import QtGui, QtWidgets, QtCore
    from PySide2.QtCore import Qt

    # Mimic PyQt6 API changes
    QtGui.QAction = QtWidgets.QAction
    QtWidgets.QApplication.exec = QtWidgets.QApplication.exec_

    # Set backend variables
    backend = 'PySide2'
    qtversion = 5

else:
    raise Exception('Selection of qt version failed.')


# Configure Matplotlib for use with Qt
import matplotlib  # noqa
try:
    matplotlib.use('QtAgg')
except ImportError:
    # In matplotlib 3.7.0 this raises ImportErrors if a previous backend
    # was already set.
    pass
try:
    # New Matplotlib no longer has qt4agg/qt5agg, but now uses qtagg for all
    import matplotlib.backends.back_qtagg as matplotlib_backend  # noqa
except ImportError:
    # https://matplotlib.org/stable/api/backend_qt_api.html
    import matplotlib.backends.backend_qt5agg as matplotlib_backend  # noqa

# Delete temporary variables
del pyqt5, pyqt6, pyside2, pyside6


# Load Gnome theme on Wayland (for icons)
if platform == 'Linux':
    icon = QtGui.QIcon.fromTheme('document-new')
    if icon.isNull():
        QtGui.QIcon.setThemeName('gnome')

# Icons with fallback for apple and windows
ICON_PATH = os.path.join(myokit.DIR_DATA, 'gui')
ICONS = {
    'document-new': 'new.png',
    'document-open': 'open.png',
    'document-save': 'save.png',
    'edit-undo': 'undo.png',
    'edit-redo': 'redo.png',
    'edit-find': 'find.png',
    'media-playback-start': 'run.png',
}
for k, v in ICONS.items():
    ICONS[k] = os.path.join(ICON_PATH, v)

# Toolbar style suitable for platform
TOOL_BUTTON_STYLE = Qt.ToolButtonStyle.ToolButtonTextUnderIcon
if platform == 'Windows':   # pragma: no linux cover
    TOOL_BUTTON_STYLE = Qt.ToolButtonStyle.ToolButtonIconOnly
elif platform == 'Darwin':  # pragma: no linux cover
    TOOL_BUTTON_STYLE = Qt.ToolButtonStyle.ToolButtonTextOnly


# Stand alone applications
class MyokitApplication(QtWidgets.QMainWindow):
    """
    Base class for Myokit applications.

    *Extends*: ``QtWidgets.QMainWindow``.
    """


def icon(name):
    """
    Returns a QtIcon created either from the theme or from one of the fallback
    icons.

    Raises a ``KeyError`` if no such icon is available.
    """
    return QtGui.QIcon.fromTheme(name, QtGui.QIcon(ICONS[name]))


def qtMonospaceFont():
    """
    Attempts to create and return a monospace font.
    """
    font = QtGui.QFont('monospace')
    if platform == 'Windows':   # pragma: no linux cover
        font.setStyleHint(QtGui.QFont.StyleHint.TypeWriter)
    else:
        font.setStyleHint(QtGui.QFont.StyleHint.Monospace)
    font.setHintingPreference(
        QtGui.QFont.HintingPreference.PreferVerticalHinting)
    return font


def run(app, *args):
    """
    Runs a Myokit gui app as a stand-alone application.

    Arguments:

    ``app``
        The application to run, specified as a class object (not an instance).
    ``*args``
        Any arguments to pass to the app's constructor.

    Example usage:

        load(myokit.gui.MyokitIDE, 'model.mmt')


    """
    # Test application class
    if not issubclass(app, MyokitApplication):
        raise ValueError(
            'Application must be specified as a type extending'
            ' MyokitApplication.')

    # Create Qt app
    a = QtWidgets.QApplication([])

    # Apply custom styling if required
    #_style_application(a)
    # Close with last window
    a.lastWindowClosed.connect(a.quit)

    # Close on Ctrl-C
    def int_signal(signum, frame):
        a.closeAllWindows()
    signal.signal(signal.SIGINT, int_signal)

    # Create app and show
    app = app(*args)
    app.show()

    # For some reason, Qt needs focus to handle the SIGINT catching...
    timer = QtCore.QTimer()
    timer.start(500)  # Flags timeout every 500ms
    timer.timeout.connect(lambda: None)

    # Wait for app to exit
    sys.exit(a.exec())

